import datetime
import os
import re
from typing import Optional, List, Tuple, Type

from flask import Flask, abort
from markdown import markdown

from .config import config
from .latex import MarkdownLatex
from ._sorters import PagesSorter, PagesSortByTime


class BlogApp(Flask):
    _title_header_regex = re.compile(r"^#\s*((\[(.*)\])|(.*))")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, template_folder=config.templates_dir, **kwargs)
        self.pages_dir = os.path.join(config.content_dir, "markdown")
        self.img_dir = config.default_img_dir
        self.css_dir = config.default_css_dir
        self.js_dir = config.default_js_dir
        self.fonts_dir = config.default_fonts_dir

        if not os.path.isdir(self.pages_dir):
            # If the `markdown` subfolder does not exist, then the whole
            # `config.content_dir` is treated as the root for markdown files.
            self.pages_dir = config.content_dir

        img_dir = os.path.join(config.content_dir, "img")
        if os.path.isdir(img_dir):
            self.img_dir = os.path.abspath(img_dir)
        else:
            self.img_dir = config.content_dir

        css_dir = os.path.join(config.content_dir, "css")
        if os.path.isdir(css_dir):
            self.css_dir = os.path.abspath(css_dir)

        js_dir = os.path.join(config.content_dir, "js")
        if os.path.isdir(js_dir):
            self.js_dir = os.path.abspath(js_dir)

        fonts_dir = os.path.join(config.content_dir, "fonts")
        if os.path.isdir(fonts_dir):
            self.fonts_dir = os.path.abspath(fonts_dir)

        templates_dir = os.path.join(config.content_dir, "templates")
        if os.path.isdir(templates_dir):
            self.template_folder = os.path.abspath(templates_dir)

    def get_page_metadata(self, page: str) -> dict:
        if not page.endswith(".md"):
            page = page + ".md"

        md_file = os.path.join(self.pages_dir, page)
        if not os.path.isfile(md_file):
            abort(404)

        metadata = {}
        with open(md_file, "r") as f:
            metadata["uri"] = "/article/" + page[:-3]

            for line in f:
                if not line:
                    continue

                if not (m := re.match(r"^\[//]: # \(([^:]+):\s*(.*)\)\s*$", line)):
                    break

                if m.group(1) == "published":
                    metadata[m.group(1)] = datetime.datetime.fromisoformat(
                        m.group(2)
                    ).date()
                else:
                    metadata[m.group(1)] = m.group(2)

        if not metadata.get("title"):
            # If the `title` header isn't available in the file,
            # infer it from the first line of the file
            with open(md_file, "r") as f:
                header = ""
                for line in f.readlines():
                    header = line
                    break

            metadata["title_inferred"] = True
            m = self._title_header_regex.search(header)
            if m:
                metadata["title"] = m.group(3) or m.group(1)
            else:
                metadata["title"] = os.path.basename(md_file)

        if not metadata.get("published"):
            # If the `published` header isn't available in the file,
            # infer it from the file's creation date
            metadata["published"] = datetime.date.fromtimestamp(
                os.stat(md_file).st_ctime
            )
            metadata["published_inferred"] = True

        return metadata

    def get_page(
        self,
        page: str,
        title: Optional[str] = None,
        skip_header: bool = False,
        skip_html_head: bool = False,
    ):
        if not page.endswith(".md"):
            page = page + ".md"

        metadata = self.get_page_metadata(page)
        # Don't duplicate the page title if it's been inferred
        if not (title or metadata.get("title_inferred")):
            title = metadata.get("title", config.title)

        with open(os.path.join(self.pages_dir, page), "r") as f:
            return render_template(
                "article.html",
                config=config,
                title=title,
                url=metadata.get("uri"),
                image=metadata.get("image"),
                description=metadata.get("description"),
                author=(
                    re.match(r"(.+?)\s+<([^>]+>)", metadata["author"])[1]
                    if "author" in metadata
                    else None
                ),
                author_email=(
                    re.match(r"(.+?)\s+<([^>]+)>", metadata["author"])[2]
                    if "author" in metadata
                    else None
                ),
                published=(
                    metadata["published"].strftime("%b %d, %Y")
                    if metadata.get("published")
                    and not metadata.get("published_inferred")
                    else None
                ),
                content=markdown(
                    f.read(),
                    extensions=["fenced_code", "codehilite", "tables", MarkdownLatex()],
                ),
                skip_header=skip_header,
                skip_html_head=skip_html_head,
            )

    def get_pages(
        self,
        with_content: bool = False,
        skip_header: bool = False,
        skip_html_head: bool = False,
        sorter: Type[PagesSorter] = PagesSortByTime,
        reverse: bool = True,
    ) -> List[Tuple[int, dict]]:
        pages_dir = app.pages_dir.rstrip("/")
        pages = [
            {
                "path": os.path.join(root[len(pages_dir) + 1 :], f),
                "folder": root[len(pages_dir) + 1 :],
                "content": (
                    self.get_page(
                        os.path.join(root, f),
                        skip_header=skip_header,
                        skip_html_head=skip_html_head,
                    )
                    if with_content
                    else ""
                ),
                **self.get_page_metadata(os.path.join(root[len(pages_dir) + 1 :], f)),
            }
            for root, _, files in os.walk(pages_dir, followlinks=True)
            for f in files
            if f.endswith(".md")
        ]

        sorter_func = sorter(pages)
        pages.sort(key=sorter_func, reverse=reverse)
        return [(i, page) for i, page in enumerate(pages)]


app = BlogApp(__name__)


from .routes import *


# vim:sw=4:ts=4:et:



import os
import sys

from .cli import get_args
from .config import init_config

arg_delim_idx = [
    i for i, arg in enumerate(sys.argv) if arg == 'madblog.uwsgi'
][0]

opts, _ = get_args(sys.argv[arg_delim_idx+1:])
config_file = os.path.join(opts.dir, 'config.yaml')
init_config(config_file=config_file, content_dir=opts.dir)

from .app import app

# For gunicorn/uWSGI compatibility
application = app


# vim:sw=4:ts=4:et:



import os
import re
from typing import Optional
from urllib.parse import urljoin

from flask import (
    jsonify,
    request,
    Response,
    send_from_directory as send_from_directory_,
    render_template,
)

from .app import app
from .config import config
from ._sorters import PagesSortByTimeGroupedByFolder


def send_from_directory(
    path: str, file: str, alternative_path: Optional[str] = None, *args, **kwargs
):
    if not os.path.exists(os.path.join(path, file)) and alternative_path:
        path = alternative_path
    return send_from_directory_(path, file, *args, **kwargs)


@app.route("/", methods=["GET"])
def home_route():
    return render_template(
        "index.html",
        pages=app.get_pages(sorter=PagesSortByTimeGroupedByFolder),
        config=config,
    )


@app.route("/img/<img>", methods=["GET"])
def img_route(img: str):
    return send_from_directory(app.img_dir, img, config.default_img_dir)


@app.route("/favicon.ico", methods=["GET"])
def favicon_route():
    return img_route("favicon.ico")


@app.route("/js/<file>", methods=["GET"])
def js_route(file: str):
    return send_from_directory(app.js_dir, file, config.default_js_dir)


@app.route("/pwabuilder-sw.js", methods=["GET"])
def pwa_builder_route():
    return send_from_directory(app.js_dir, "pwabuilder-sw.js", config.default_js_dir)


@app.route("/pwabuilder-sw-register.js", methods=["GET"])
def pwa_builder_register_route():
    return send_from_directory(
        app.js_dir, "pwabuilder-sw-register.js", config.default_js_dir
    )


@app.route("/css/<style>", methods=["GET"])
def css_route(style: str):
    return send_from_directory(app.css_dir, style, config.default_css_dir)


@app.route("/fonts/<file>", methods=["GET"])
def fonts_route(file: str):
    return send_from_directory(app.fonts_dir, file, config.default_fonts_dir)


@app.route("/manifest.json", methods=["GET"])
def manifest_route():
    # If there is a manifest.json in the content directory, use it
    manifest_file = os.path.join(config.content_dir, "manifest.json")
    if os.path.isfile(manifest_file):
        return send_from_directory(config.content_dir, "manifest.json")

    # Otherwise, generate a default manifest.json
    return jsonify(
        {
            "name": config.title,
            "short_name": config.title,
            "icons": [
                {"src": "/img/icon-48.png", "sizes": "48x48", "type": "image/png"},
                {"src": "/img/icon-72.png", "sizes": "72x72", "type": "image/png"},
                {"src": "/img/icon-96.png", "sizes": "96x96", "type": "image/png"},
                {"src": "/img/icon-144.png", "sizes": "144x144", "type": "image/png"},
                {"src": "/img/icon-168.png", "sizes": "168x168", "type": "image/png"},
                {"src": "/img/icon-192.png", "sizes": "192x192", "type": "image/png"},
                {"src": "/img/icon-256.png", "sizes": "256x256", "type": "image/png"},
                {"src": "/img/icon-512.png", "sizes": "512x512", "type": "image/png"},
            ],
            "gcm_sender_id": "",
            "gcm_user_visible_only": True,
            "start_url": "/",
            "permissions": ["gcm"],
            "scope": "",
            "orientation": "portrait",
            "display": "standalone",
            "theme_color": "#000000",
            "background_color": "#ffffff",
        }
    )


@app.route("/article/<path:path>/<article>", methods=["GET"])
def article_with_path_route(path: str, article: str):
    return app.get_page(os.path.join(path, article))


@app.route("/article/<article>", methods=["GET"])
def article_route(article: str):
    return article_with_path_route("", article)


@app.route("/rss", methods=["GET"])
def rss_route():
    short_description = "short" in request.args or config.short_feed
    pages = app.get_pages(
        with_content=not short_description,
        skip_header=True,
        skip_html_head=True,
    )

    return Response(
        """<?xml version="1.0" encoding="UTF-8" ?>
<rss version="2.0" xmlns:media="http://search.yahoo.com/mrss/">
    <channel>
        <title>{title}</title>
        <link>{link}</link>
        <description>{description}</description>
        <category>{categories}</category>
        <image>
            <url>{link}/img/icon.png</url>
            <title>{title}</title>
            <link>{link}</link>
        </image>
        <pubDate>{last_pub_date}</pubDate>
        <language>{language}</language>

        {items}
    </channel>
</rss>""".format(
            title=config.title,
            description=config.description,
            link=config.link,
            categories=",".join(config.categories),
            language=config.language,
            last_pub_date=(
                pages[0][1]["published"].strftime("%a, %d %b %Y %H:%M:%S GMT")
                if pages
                else ""
            ),
            items="\n\n".join(
                [
                    (
                        """
            <item>
                <title>{title}</title>
                <link>{base_link}{link}</link>
                <pubDate>{published}</pubDate>
                <description><![CDATA[{content}]]></description>
                <media:content medium="image" url="{image}" width="200" height="150" />
            </item>
                    """
                    ).format(
                        base_link=config.link,
                        title=page.get("title", "[No Title]"),
                        link=page.get("uri", ""),
                        published=(
                            page["published"].strftime("%a, %d %b %Y %H:%M:%S GMT")
                            if "published" in page
                            else ""
                        ),
                        content=(
                            page.get("description", "")
                            if short_description
                            else page.get("content", "")
                        ),
                        image=(
                            urljoin(config.link, page["image"])
                            if page.get("image")
                            and not re.search(r"^https?://", page["image"])
                            else page.get("image", "")
                        ),
                    )
                    for _, page in pages
                ]
            ),
        ),
        mimetype="application/xml",
    )


# vim:sw=4:ts=4:et:



import os
from typing import List
import yaml

from dataclasses import dataclass, field


@dataclass
class Config:
    title = "Blog"
    description = ""
    link = "/"
    home_link = "/"
    language = "en-US"
    logo = "/img/icon.png"
    header = True
    content_dir = "."
    categories: List[str] = field(default_factory=list)
    short_feed = False

    basedir = os.path.abspath(os.path.dirname(__file__))
    templates_dir = os.path.join(basedir, "templates")
    static_dir = os.path.join(basedir, "static")
    default_css_dir = os.path.join(static_dir, "css")
    default_js_dir = os.path.join(static_dir, "js")
    default_fonts_dir = os.path.join(static_dir, "fonts")
    default_img_dir = os.path.join(static_dir, "img")


config = Config()


def init_config(content_dir=".", config_file="config.yaml"):
    cfg = {}
    config.content_dir = content_dir

    if os.path.isfile(config_file):
        with open(config_file, "r") as f:
            cfg = yaml.safe_load(f)

    if cfg.get("title"):
        config.title = cfg["title"]
    if cfg.get("description"):
        config.description = cfg["description"]
    if cfg.get("link"):
        config.link = cfg["link"]
    if cfg.get("home_link"):
        config.home_link = cfg["home_link"]
    if cfg.get("logo") is not None:
        config.logo = cfg["logo"]
    if cfg.get("language"):
        config.language = cfg["language"]
    if cfg.get("header") is False:
        config.header = False
    if cfg.get("short_feed"):
        config.short_feed = True

    config.categories = cfg.get("categories", [])


# vim:sw=4:ts=4:et:
